from pylabnet.scripts.pid import PID
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.gui.pyqt.gui_handler import GUIHandler
from pylabnet.utils.helper_methods import unpack_launcher, create_server
from pylabnet.utils.logging.logger import LogClient

import numpy as np
import time
import copy
import pickle


# Static methods

def generate_widgets():
    """Static method to return systematically named gui widgets for 4ch wavemeter monitor"""

    graphs, legends, numbers, booleans, labels, events = [], [], [], [], [], []
    for i in range(4):
        graphs.append('graph_widget_' + str(i + 1))
        legends.append('legend_widget_' + str(i + 1))
        numbers.append('number_widget_' + str(i + 1))
        booleans.append('boolean_widget_' + str(i + 1))
        labels.append('label_' + str(i + 1))
        events.append('event_button_' + str(i + 1))
    for i in range(4, 8):
        numbers.append('number_widget_' + str(i + 1))
        booleans.append('boolean_widget_' + str(i + 1))
        labels.append('label_' + str(i + 1))
    return graphs, legends, numbers, booleans, labels, events


# Core objects

class WlmMonitor:
    """ A script class for monitoring and locking lasers based on the wavemeter """

    # Assign widget names based on .gui file.
    (_graph_widgets,
     _legend_widgets,
     _number_widgets,
     _boolean_widgets,
     _label_widgets,
     _event_widgets) = generate_widgets()

    def __init__(self, wlm_client, gui_client, logger_client, ao_clients=None, display_pts=5000, threshold=0.0002):
        """ Instantiates WlmMonitor script object for monitoring wavemeter

        :param wlm_client: (obj) instance of wavemeter client
        :param gui_client: (obj) instance of GUI client.
        :param logger_client: (obj) instance of logger client.
        :param ao_clients: (dict, optional) dictionary of AO client objects with keys to identify. Exmaple:
            {'ni_usb_1': nidaqmx_usb_client_1, 'ni_usb_2': nidaqmx_usb_client_2, 'ni_pxi_multi': nidaqmx_pxi_client}
        :param display_pts: (int, optional) number of points to display on plot
        :param threshold: (float, optional) threshold in THz for lock error signal
        """
        self.channels = []
        self.wlm_client = wlm_client
        self.ao_clients = ao_clients
        self.display_pts = display_pts
        self.threshold = threshold

        # Instanciate gui handler
        self.gui_handler = GUIHandler(gui_client, logger_client)

    def set_parameters(self, channel_params):
        """ Instantiates new channel objects with given parameters and assigns them to the WlmMonitor

        Note: parameters for a channel that has already been assigned can be set or updated later using the
        update_parameters() method via an update client in a separate python process.

        :param channel_params: (list) of dictionaries containing all parameters. Example of full parameter set:
            {'channel': 1, 'name': 'Velocity', 'setpoint': 406.7, 'lock':True, 'memory': 20,
             'PID': {'p': 1, 'i': 0.1, 'd': 0}, 'AO': {'client':'nidaqmx_client', 'channel': 'ao1'},
             'voltage_monitor': True}

            In more detail:
            - 'channel': should be from 1-8 for the High-Finesse Wavemeter (with switch) and should ALWAYS be provided,
                as a reference so that we know which channel to assign all the other parameters to
            - 'name': a string that can just be provided once and is used as a user-friendly name for the channel.
                Initializes to 'Channel X' where X is a random integer if not provided
            - 'setpoint': setpoint for this channel. If not provided, or if None, the setpoint is not plotted/tracked
            - 'lock': boolean that tells us whether or not to turn on the lock. Ignored if setpoint is None. Default is
                False.
            - 'memory': Number of points for integral memory of PID (history of the integral). Default is 20.
            - 'PID': dict containing PID parameters. Uses the pylabnet.scripts.PID module. By default instantiates the
                default PID() object.
            - 'AO': dict containing two elements: 'client' which is a string that is the name of the AO client to use
                for locking. This should match up with a key in self.ao_clients. 'channel'is an identifier for which
                analog output to use for this channel. By default it is set to None and no active locking is performed
            - 'voltage monitor': boolean which tells us whether to also plot/track the AO voltage + setpoint error on
                the plot below the laser graph. Default is False
        """

        # Check if it is only a single channel
        if type(channel_params) is dict:
            channel_params = [channel_params]

        # Initialize each channel individually
        for channel_param_set in channel_params:
            self.channels.append(Channel(channel_param_set, self.ao_clients))

    def update_parameters(self, parameters):
        """ Updates only the parameters given. Can be used in the middle of the script operation via an update client.

        :param parameters: (list) list of dictionaries, see set_parameters() for details
        """

        for parameter in parameters:

            # Make sure a channel is given
            if 'channel' in parameter:

                # Check if it is a channel that is already logged
                channel_list = self._get_channels()
                if parameter['channel'] in channel_list:

                    # Find index of the desired channel
                    channel = self.channels[channel_list.index(parameter['channel'])]

                    # Set all other parameters for this channel
                    if 'name' in parameter:
                        channel.name = parameter['name']

                    if 'setpoint' in parameter:

                        # If the setpoint didn't exist and now exists, we need to update and add a curve
                        if channel.setpoint is None and parameter['setpoint'] is not None:
                            self.gui_handler.assign_curve(
                                plot_label=channel.name,
                                curve_label=channel.setpoint_name
                            )
                            channel.setpoint = parameter['setpoint']
                            channel.pid.set_parameters(setpoint=channel.setpoint)
                            channel.initialize_sp_data()

                        # If the setpoint existed and is now removed, delete the curve item
                        elif channel.setpoint is not None and parameter['setpoint'] is None:
                            self.gui_handler.remove_curve(
                                plot_label=channel.name,
                                curve_label=channel.setpoint_name
                            )
                            channel.setpoint = None
                            channel.pid.set_parameters(setpoint=0)

                        # Otherwise just update the setpoint normally
                        else:
                            channel.setpoint = parameter['setpoint']

                            # Mark that we should override GUI setpoint, since it has been updated by the script
                            channel.setpoint_override = True

                    if 'lock' in parameter:
                        channel.lock = parameter['lock']

                        # Mark that we should override the GUI lock since it has been updated by the script
                        channel.lock_override = True

                    if 'memory' in parameter:
                        channel.memory = parameter['memory']
                    if 'PID' in parameter:
                        channel.pid.set_parameters(
                            p=parameter['PID']['p'],
                            i=parameter['PID']['i'],
                            d=parameter['PID']['d'],
                        )

                    # Ignore AO requests if clients have not been assigned
                    if 'AO' in parameter and self.ao_clients is not None:

                        # Convert AO from string to object using lookup
                        try:
                            channel.ao = {
                                'client': self.ao_clients[parameter['AO']['client']],
                                'channel': parameter['AO']['channel']
                            }

                        # Handle case where the AO client does not exist
                        except KeyError:
                            channel.ao = None

                # Otherwise, it is a new channel so we should add it
                else:
                    self.channels.append(Channel(parameter))
                    self._initialize_channel(
                        index=len(self.channels) - 1,
                        channel=self.channels[-1]
                    )

    def initialize_channels(self):
        """Initializes all channels and outputs to the GUI"""

        for index, channel in enumerate(self.channels):
            self._initialize_channel(index, channel)

    def clear_channel(self, channel):
        """ Clears the plot output for this channel

        :param channel: channel number to clear
        """

        # Get the relevant channel
        channel_list = self._get_channels()

        try:
            channel = self.channels[channel_list.index(channel)]
            channel.initialize(channel.data[-1])

        # If the channel isn't monitored
        except KeyError:
            pass

    def run(self):
        """Runs the WlmMonitor continuously

        Can be stopped using the pause() method
        """

        # Mark running flag
        self.is_running = True
        while self.is_running:
            self._get_gui_data()
            self._update_channels()
            time.sleep(0.02)

    def pause(self):
        """Pauses the wavemeter monitor"""
        self.is_running = False
        self.is_paused = True

    def resume(self):
        """Resumes the wavemeter monitor when paused"""
        self.is_paused = False

    def reconnect_gui(self):
        """ Reconnects to the GUI

        Should be called if the GUI connection has been lost, once a new GUI client with the same access parameters has
        been reinstantiated
        """
        self.gui_handler.gui_reconnect = True

    def zero_voltage(self, channel):
        """ Zeros the output voltage for this channel

        :param channel: (int) channel number to zero voltage of
        """

        # Get the relevant channel
        channel_list = self._get_channels()

        try:
            channel = self.channels[channel_list.index(channel)]
            channel.zero_voltage()

        # If the channel isn't monitored
        except KeyError:
            pass

    # Technical methods

    def _initialize_channel(self, index, channel):
        """Initializes a channel and outputs to the GUI

        Should only be called in the beginning of channel use to assign physical GUI widgets
        """

        # Get wavelength and initialize data arrays
        channel.initialize(
            wavelength=self.wlm_client.get_wavelength(channel.number),
            display_pts=self.display_pts
        )

        # Calculate indices for various gui widget defaults. This is GUI specific
        if channel.voltage is None:
            plot_multiplier = 2
            scalar_multiplier = 4
        else:
            plot_multiplier = 1
            scalar_multiplier = 2

        # First, clear the plot in case it was in use by some previous process
        self.gui_handler.clear_plot(
            plot_widget=self._graph_widgets[plot_multiplier * (index + channel.plot_widget_offset)]
        )

        # Clear voltage if relevant
        if channel.voltage is not None:

            # First clear the plot
            self.gui_handler.clear_plot(
                plot_widget=self._graph_widgets[plot_multiplier * (index + channel.plot_widget_offset) + 1]
            )

        # Assign GUI + curves
        self.gui_handler.assign_plot(
            plot_widget=self._graph_widgets[plot_multiplier * (index + channel.plot_widget_offset)],
            plot_label=channel.name,
            legend_widget=self._legend_widgets[plot_multiplier * (index + channel.plot_widget_offset)]
        )
        self.gui_handler.assign_curve(
            plot_label=channel.name,
            curve_label=channel.curve_name
        )

        # Numeric label for laser frequency
        self.gui_handler.assign_scalar(
            scalar_widget=self._number_widgets[scalar_multiplier * (index + channel.plot_widget_offset)],
            scalar_label=channel.name
        )

        # Only initialize setpoint plot if it is provided, since we don't want to clutter the screen
        if channel.setpoint is not None:
            self.gui_handler.assign_curve(
                plot_label=channel.name,
                curve_label=channel.setpoint_name
            )

        # Numeric label for setpoint
        self.gui_handler.assign_scalar(
            scalar_widget=self._number_widgets[scalar_multiplier * (index + channel.plot_widget_offset) + 1],
            scalar_label=channel.setpoint_name
        )

        # Assign lock and error boolean widgets
        self.gui_handler.assign_scalar(
            scalar_widget=self._boolean_widgets[scalar_multiplier * (index + channel.plot_widget_offset)],
            scalar_label=channel.lock_name
        )
        self.gui_handler.assign_scalar(
            scalar_widget=self._boolean_widgets[scalar_multiplier * (index + channel.plot_widget_offset) + 1],
            scalar_label=channel.error_name
        )

        # Assign pushbutton for clearing data
        self.gui_handler.assign_event_button(
            event_widget=self._event_widgets[plot_multiplier * (index + channel.plot_widget_offset)],
            event_label=channel.name
        )

        # Assign voltage if relevant
        if channel.voltage is not None:

            # Now reassign it
            self.gui_handler.assign_plot(
                plot_widget=self._graph_widgets[plot_multiplier * (index + channel.plot_widget_offset) + 1],
                plot_label=channel.aux_name,
                legend_widget=self._legend_widgets[plot_multiplier * (index + channel.plot_widget_offset) + 1]
            )
            self.gui_handler.assign_curve(
                plot_label=channel.aux_name,
                curve_label=channel.voltage_curve
            )
            self.gui_handler.assign_curve(
                plot_label=channel.aux_name,
                curve_label=channel.error_curve
            )

            # Display scalars as well
            self.gui_handler.assign_scalar(
                scalar_widget=self._number_widgets[scalar_multiplier * (index + channel.plot_widget_offset) + 2],
                scalar_label=channel.voltage_curve
            )
            self.gui_handler.assign_scalar(
                scalar_widget=self._number_widgets[scalar_multiplier * (index + channel.plot_widget_offset) + 3],
                scalar_label=channel.error_curve
            )

            # Change label text for voltage
            self.gui_handler.assign_label(
                label_widget=self._label_widgets[scalar_multiplier * (index + channel.plot_widget_offset) + 2],
                label_label=channel.voltage_curve
            )
            self.gui_handler.assign_label(
                label_widget=self._label_widgets[scalar_multiplier * (index + channel.plot_widget_offset) + 3],
                label_label=channel.error_curve
            )

    def _update_channels(self):
        """ Updates all channels + displays

        Called continuously inside run() method to refresh WLM data and output on GUI
        """

        for channel in self.channels:

            # Update data with the new wavelength
            channel.update(self.wlm_client.get_wavelength(channel.number))

            # Try to update plots if we have a GUI connected
            if self.gui_handler.gui_connected:

                self.gui_handler.set_curve_data(
                    data=channel.data,
                    plot_label=channel.name,
                    curve_label=channel.curve_name
                )
                self.gui_handler.set_scalar(
                    value=channel.data[-1],
                    scalar_label=channel.name
                )

                # Update setpoints if necessary
                if channel.setpoint is not None:
                    self.gui_handler.set_curve_data(
                        data=channel.sp_data,
                        plot_label=channel.name,
                        curve_label=channel.setpoint_name
                    )

                    # Update the setpoint to GUI directly if it has been changed
                    if channel.setpoint_override:

                        # Tell GUI to pull data provided by script and overwrite direct GUI input
                        self.gui_handler.activate_scalar(
                            scalar_label=channel.setpoint_name
                        )
                        self.gui_handler.set_scalar(
                            value=channel.setpoint,
                            scalar_label=channel.setpoint_name
                        )

                    # Otherwise, tell GUI to stop updating from script and accept direct GUI input
                    else:
                        self.gui_handler.deactivate_scalar(
                            scalar_label=channel.setpoint_name
                        )

                # Set lock and error booleans
                if channel.setpoint is not None:

                    # If the lock has been updated, override the GUI
                    if channel.lock_override:
                        self.gui_handler.activate_scalar(
                            scalar_label=channel.lock_name
                        )
                        self.gui_handler.set_scalar(
                            value=channel.lock,
                            scalar_label=channel.lock_name
                        )

                    # Otherwise, enable GUI input
                    else:
                        self.gui_handler.deactivate_scalar(
                            scalar_label=channel.lock_name
                        )

                    # Set the error boolean (true if the lock is active and we are outside the error threshold)
                    if channel.lock and np.abs(channel.data[-1] - channel.setpoint) > self.threshold:
                        self.gui_handler.set_scalar(
                            value=True,
                            scalar_label=channel.error_name
                        )
                    else:
                        self.gui_handler.set_scalar(
                            value=False,
                            scalar_label=channel.error_name
                        )

                # Otherwise just set everything false, since we don't have a setpoint
                else:
                    self.gui_handler.set_scalar(
                        value=False,
                        scalar_label=channel.lock_name
                    )
                    self.gui_handler.set_scalar(
                        value=False,
                        scalar_label=channel.error_name
                    )

                # Now update lock + voltage plots + scalars if relevant
                if channel.voltage is not None:
                    self.gui_handler.set_curve_data(
                        data=channel.voltage,
                        plot_label=channel.aux_name,
                        curve_label=channel.voltage_curve
                    )
                    self.gui_handler.set_curve_data(
                        data=channel.error,
                        plot_label=channel.aux_name,
                        curve_label=channel.error_curve
                    )
                    self.gui_handler.set_scalar(
                        value=channel.voltage[-1],
                        scalar_label=channel.voltage_curve
                    )
                    self.gui_handler.set_scalar(
                        value=channel.error[-1],
                        scalar_label=channel.error_curve
                    )

                    # Update labels, if desired
                    if not channel.labels_updated:
                        self.gui_handler.set_label(
                            text='Voltage',
                            label_label=channel.voltage_curve
                        )
                        self.gui_handler.set_label(
                            text='Lock Error',
                            label_label=channel.error_curve
                        )
                        channel.label_updated = True

            # If GUI is not connected, check if we should try reconnecting to the GUI
            elif  self.gui_handler.gui_reconnect:

                self.gui_handler.gui_connected = True
                # self.gui_handler.connect() # Commented by CK, unsure what that does.

                # Reinitialize channels to new GUI
                self.initialize_channels()

                # Manually update the GUI for a while so we don't change the setpoints
                channel.setpoint_override = True
                channel.lock_override = True
                self.gui_handler.activate_scalar(scalar_label=channel.setpoint_name)
                self.gui_handler.activate_scalar(scalar_label=channel.lock_name)
                self.gui_handler.set_scalar(
                    value=channel.setpoint,
                    scalar_label=channel.setpoint_name
                )
                self.gui_handler.set_scalar(
                    value=channel.lock,
                    scalar_label=channel.lock_name
                )
                self._update_channels()
                self.gui_handler.gui_reconnect = False

    def _get_gui_data(self):
        """ Updates setpoint and lock parameters with data pulled from GUI

        Does not overwrite the script setpoints and locks, but stores the GUI values for comparison based on context.
        See Channel.update() method for behavior on how script chooses whether to use internal values or GUI values
        """
        for channel in self.channels:

            # Pull the current value from the GUI
            if channel.setpoint is not None and self.gui_handler.gui_connected:
                channel.gui_setpoint = self.gui_handler.get_scalar(scalar_label=channel.setpoint_name)
                channel.gui_lock = self.gui_handler.get_scalar(scalar_label=channel.lock_name)

            if self.gui_handler.gui_connected and self.gui_handler.was_button_pressed(event_label=channel.name):
                self.clear_channel(channel=channel.number)

    def _get_channels(self):
        """ Returns all active channel numbers

        Usually used for checking whether a newly input channel has already been assigned to the script

        :return: (list) all active channel numbers
        """

        channel_list = []
        for channel in self.channels:
            channel_list.append(channel.number)
        return channel_list


class Service(ServiceBase):
    """ A service to enable external updating of WlmMonitor parameters """

    def exposed_update_parameters(self, params_pickle):

        params = pickle.loads(params_pickle)
        return self._module.update_parameters(params)

    def exposed_clear_channel(self, channel):
        return self._module.clear_channel(channel)

    def exposed_reconnect_gui(self):
        return self._module.reconnect_gui()

    def exposed_zero_voltage(self, channel):
        return self._module.zero_voltage(channel)

    def exposed_pause(self):

        if isinstance(self._module, list):
            for module in self._module:
                module.pause()
            return 0

        else:
            return self._module.pause()

    def exposed_resume(self):
        return self._module.resume()


class Client(ClientBase):

    def update_parameters(self, params):

        params_pickle = pickle.dumps(params)
        return self._service.exposed_update_parameters(params_pickle)

    def clear_channel(self, channel):
        return self._service.exposed_clear_channel(channel)

    def zero_voltage(self, channel):
        return self._service.exposed_zero_voltage(channel)

    def reconnect_gui(self):
        return self._service.exposed_reconnect_gui()

    def pause(self):

        return self._service.exposed_pause()

    def resume(self):
        return self._service.exposed_resume()


class Channel:
    """Object containing all information regarding a single wavemeter channel"""

    # Set acceptable voltage range for laser modulation
    _min_voltage = -10
    _max_voltage = 10

    # Gain multiplier for PID
    _gain = 1000

    def __init__(self, channel_params, ao_clients=None):
        """
        Initializes all parameters given, sets others to default. Also sets up some defaults + placeholders for data

        :param channel_params: (dict) Dictionary of channel parameters (see WlmMonitor.set_parameters() for details)
        :param ao_clients: (dict, optional) Dictionary containing AO clients tying a keyname string to the actual client
        """

        # Set channel parameters to default values
        self.ao_clients = ao_clients
        self.ao = None  # Dict with client name and channel for AO to use
        self.voltage = None  # Array of voltage values for AO, used for plotting/monitoring voltage
        self.current_voltage = 0
        self.setpoint = None
        self.lock = False
        self.error = None  # Array of error values, used for plotting/monitoring lock error
        self.labels_updated = False  # Flag to check if we have updated all labels
        self.setpoint_override = True  # Flag to check if setpoint has been updated + GUI should be overridden
        self.lock_override = True  # Flag to check if lock has been updated + GUI should be overridden
        self.gui_setpoint = 0  # Current GUI setpoint
        self.gui_lock = False  # Current GUI lock boolean
        self.prev_gui_lock = None  # Previous GUI lock boolean
        self.prev_gui_setpoint = None  # Previous GUI setpoint

        # Set all relevant parameters to default values
        self._overwrite_parameters(channel_params)

        # Initialize relevant placeholders
        self.data = np.array([])
        self.sp_data = np.array([])

    def initialize(self, wavelength, display_pts=5000):
        """
        Initializes the channel based on the current wavelength

        :param wavelength: current wavelength
        :param display_pts: number of points to display on the plot
        """

        self.data = np.ones(display_pts) * wavelength

        if self.setpoint is not None:
            self.sp_data = np.ones(display_pts) * self.setpoint
            self.setpoint_override = True

        self.lock_override = True

        # Initialize voltage and error
        if self.voltage is not None:
            self.voltage = np.ones(display_pts) * self.current_voltage

            # Check that setpoint is reasonable, otherwise set error to 0
            if self.setpoint is None:
                self.error = np.zeros(display_pts)
            else:
                self.error = np.ones(display_pts) * (wavelength - self.setpoint)

    def initialize_sp_data(self, display_pts=5000):
        if self.setpoint is not None:
            self.sp_data = np.ones(display_pts) * self.setpoint

    def update(self, wavelength):
        """
        Updates the data, setpoints, and all locks

        :param wavelength: (float) current wavelength
        """

        self.data = np.append(self.data[1:], wavelength)

        if self.setpoint is not None:

            # Pick which setpoint to use
            # If the setpoint override is on, this means we need to try and set the GUI value to self.setpoint
            if self.setpoint_override:

                # Check if the GUI has actually caught up
                if self.setpoint == self.gui_setpoint:
                    self.setpoint_override = False

                # Otherwise, the GUI still hasn't implemented the setpoint prescribed by update_parameters()

            # If setpoint override is off, this means the GUI caught up to our last update_parameters() call, and we
            # should refrain from updating the value until we get a new value from the GUI
            else:

                # Check if the GUI has changed, and if so, update the setpoint in the script to match
                if self.gui_setpoint != self.prev_gui_setpoint:
                    self.setpoint = copy.deepcopy(self.gui_setpoint)

                # Otherwise the GUI is static AND parameters haven't been updated so we don't change the setpoint at all

            # Store the latest GUI setpoint
            self.prev_gui_setpoint = copy.deepcopy(self.gui_setpoint)
            self.sp_data = np.append(self.sp_data[1:], self.setpoint)

        # Now deal with PID stuff
        self.pid.set_parameters(setpoint=0 if self.setpoint is None else self.setpoint)

        # Implement lock
        # Set process variable
        self.pid.set_pv(pv=self.data[len(self.data) - self.memory:])
        # Set control variable
        self.pid.set_cv()

        # See logic for setpoint above
        if self.lock_override:
            if self.lock == self.gui_lock:
                self.lock_override = False
        else:
            if self.gui_lock != self.prev_gui_lock:
                self.lock = copy.deepcopy(self.gui_lock)

        self.prev_gui_lock = copy.deepcopy(self.gui_lock)

        if self.lock:
            try:
                if self.ao is not None:
                    if self._min_voltage <= self.current_voltage + self.pid.cv * self._gain <= self._max_voltage:
                        self.current_voltage += self.pid.cv * self._gain
                    elif self.current_voltage + self.pid.cv * self._gain < self._min_voltage:
                        self.current_voltage = self._min_voltage
                    else:
                        self.current_voltage = self._max_voltage
                    self.ao['client'].set_ao_voltage(
                        ao_channel=self.ao['channel'],
                        voltages=[self.current_voltage]
                    )
            except EOFError:
                self.ao = None

        # Update voltage and error data
        if self.voltage is not None:
            self.voltage = np.append(self.voltage[1:], self.current_voltage)
            self.error = np.append(self.error[1:], self.pid.error * self._gain)
        else:
            print('Voltage is None')

    def zero_voltage(self):
        """Zeros the voltage (if applicable)"""

        try:
            if self.ao is not None:
                self.ao['client'].set_ao_voltage(
                    ao_channel=self.ao['channel'],
                    voltages=[0]
                )
                self.current_voltage = 0
        except EOFError:
            self.ao = None

    def _overwrite_parameters(self, channel_params):
        """ Sets all internal channel parameters to input

        If parameters are not given, they are overwritten to defaults - see implementation below, as well as the
        WlmMonitor.set_parameters() docstring for default details

        :param channel_params: (dict) dictionary containing all parameters. See WlmMonitor.set_parameters() for details
        """

        # Initialize all given attributes, otherwise initialize defaults
        if 'channel' in channel_params:
            self.number = channel_params['channel']
        else:

            # By default use channel 1
            self.number = 1

        if 'name' in channel_params:
            self.name = channel_params['name']
        else:

            # Initialize some random channel name if not given
            self.name = 'Channel ' + str(np.random.randint(1000000))
        self.curve_name = self.name + ' Frequency'  # Name used for identifying the frequency Curve object
        self.lock_name = self.name + ' Lock'  # Name used for identifying lock Scalar object
        self.error_name = self.name + ' Error'  # Name used for identifying error Scalar object

        if 'setpoint' in channel_params:
            self.setpoint = channel_params['setpoint']
        else:
            self.setpoint = None
        self.setpoint_name = self.name + ' Setpoint'  # Name used for identifying setpoint Curve object

        if 'lock' in channel_params:
            self.lock = channel_params['lock']
        else:
            self.lock = False

        if 'memory' in channel_params:
            self.memory = channel_params['memory']
        else:
            self.memory = 20

        if 'PID' in channel_params:
            self.pid = PID(
                p=channel_params['PID']['p'],
                i=channel_params['PID']['i'],
                d=channel_params['PID']['d'],
                memory=self.memory,
                setpoint=0 if self.setpoint is None else self.setpoint
            )
        else:

            # Just initialize a default PID module
            self.pid = PID()

        if 'AO' in channel_params and self.ao_clients is not None:

            # Convert AO from string to object using lookup
            try:
                self.ao = {
                    'client': self.ao_clients[channel_params['AO']['client']],
                    'channel': channel_params['AO']['channel']
                }
            except KeyError:
                # Alert the user that AO initialization failed
                self.ao = None
                print('Failed to initialize AO for Channel {}'.format(self.number))

        # If AO is not given just leave it as None
        else:
            self.ao = None

        # Configure voltage monitor arrays if desired
        if 'voltage_monitor' in channel_params:
            if channel_params['voltage_monitor']:
                self.voltage = np.array([])
                self.error = np.array([])
                self.aux_name = '{} Auxiliary Monitor'.format(self.name)
                self.voltage_curve = '{} Voltage'.format(self.name)
                self.error_curve = '{} Lock Error'.format(self.name)
            else:
                self.voltage = None

        # Set voltage to None - this will be used later on to check if we should be monitoring
        else:
            self.voltage = None

        # If we specify a specific widget to use, configure an offset for plot widget loading
        if 'plot_widget' in channel_params:
            self.plot_widget_offset = channel_params['plot_widget']

        # Otherwise just use the default widget (populates in order)
        else:
            self.plot_widget_offset = 0


def launch(**kwargs):
    """ Launches the WLM monitor + lock script """

    logger, loghost, logport, clients, guis, params = unpack_launcher(**kwargs)

    wavemeter_client = clients['high_finesse_ws7']
    ao_client = clients['nidaqmx_wlm']
    gui_client = guis['wavemeter_monitor']

    # Instantiate Monitor script
    wlm_monitor = WlmMonitor(
        wlm_client=wavemeter_client,
        gui_client=gui_client,
        ao_clients={'cDAQ1': ao_client},
        logger_client=logger
    )

    # Instantiate pause+update service & connect to logger
    log_client_update = LogClient(
        host=loghost,
        port=logport,
        module_tag='wavemeter_update_server'
    )
    update_service = Service()
    update_service.assign_module(module=wlm_monitor)
    update_service.assign_logger(logger=log_client_update)
    update_server = create_server(update_service, logger)[0]
    update_server.start()

    if params is None:
        params = dict(channel_params=[dict(
            channel=1,
            name="Velocity",
            AO=dict(client='cDAQ1', channel='ao0'),
            PID=dict(p=0.15, i=0.01, d=0),
            memory=100,
            voltage_monitor=True
            )])

    # Set parameters
    # Can use these as default parameters for loading up the monitor initially. New parameters can be input afterwards
    # using an update client
    wlm_monitor.set_parameters(**params)

    # Initialize channels to GUI
    wlm_monitor.initialize_channels()

    # Run continuously
    # Note that the actual operation inside run() can be paused using the update server
    while True:

        # Make sure the WlmMonitor is not paused, otherwise run it
        if not wlm_monitor.gui_handler.is_paused:
            wlm_monitor.run()



""" Initializes the logger + graphical display for launching pylabnet scripts """

import sys
import socket
import os
import time
import subprocess
from io import StringIO
import re
from pylabnet.utils.logging.logger import LogService
from PyQt5 import QtWidgets, QtGui, QtCore

from pylabnet.utils.logging.logger import LogService
from pylabnet.network.core.generic_server import GenericServer
from pylabnet.gui.pyqt.external_gui import Window
from pylabnet.network.client_server.external_gui import Service, Client
from pylabnet.utils.logging.logger import LogClient
from pylabnet.utils.helper_methods import dict_to_str, remove_spaces, create_server, show_console, hide_console


if hasattr(QtCore.Qt, 'AA_EnableHighDpiScaling'):
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_EnableHighDpiScaling, True)

if hasattr(QtCore.Qt, 'AA_UseHighDpiPixmaps'):
    QtWidgets.QApplication.setAttribute(QtCore.Qt.AA_UseHighDpiPixmaps, True)


class Controller:
    """ Class for log system controller """

    LOGGER_UI = 'logger_remote'
    GUI_NAME = 'logger_GUI'

    # When kept as None, random port numbers will be used
    # use these values to override and set manual port numbers if desired
    LOG_PORT = None
    GUI_PORT = None

    def __init__(self, proxy=False):
        """ Initializes launch control GUI """

        self.log_service = None
        self.log_server = None
        self.gui_client = None
        self.gui_logger = None
        self.gui_service = None
        self.log_port = self.LOG_PORT
        self.gui_port = self.GUI_PORT
        self.gui_server = None
        self.client_list = {}
        self.port_list = {}
        self.script_list = {}
        self.client_data = {}
        self.disconnection = False
        self.debug = False
        self.debug_level = None
        try:
            if sys.argv[1] == '-p' or proxy:
                self.proxy = True
            else:
                self.proxy = False
        except IndexError:
            if proxy:
                self.proxy = True
            else:
                self.proxy = False
        self.host = socket.gethostbyname(socket.gethostname())
        self.update_index = 0

        # Find logger if applicable
        if self.proxy:
            try:
                self.host = sys.argv[2]
            except IndexError:
                show_console()
                self.host = input('Please enter the master Launch Control IP address:\n>> ')
            show_console()
            self.log_port = int(input('Please enter the master Logger Port:\n>> '))
            self.gui_port = int(input('Please enter the master GUI Port:\n>> '))
            hide_console()

        sys.stdout = StringIO()

        # Instantiate GUI application
        self.app = QtWidgets.QApplication(sys.argv)
        self.main_window = Window(self.app, gui_template=self.LOGGER_UI)

    def start_gui_server(self):
        """ Starts the launch controller GUI server, or connects to the server and updates GUI"""

        module_str = ''
        if self.proxy:
            module_str = '_proxy'
        # connect to the logger
        self.gui_logger = LogClient(
            host=self.host,
            port=self.log_port,
            module_tag=self.GUI_NAME+module_str,
            ui=self.LOGGER_UI
        )

        gui_str = ''
        if self.proxy:
            gui_str = 'Master '

            # Connect to the GUI server
            try:
                self.gui_client = Client(host=self.host, port=self.gui_port)
            except ConnectionRefusedError:
                self.gui_logger.error(f'Failed to connect to GUI Server with IP address: {self.host}, '
                                      f'Port: {self:gui_port}')
                raise

            # Now update GUI to mirror clients
            self._copy_master()

            # Get the latest update index
            buffer = self.gui_client.get_text('buffer')
            self.update_index = int(re.findall(r'\d+', re.findall(r'!~\d+~!', buffer)[-1])[0])

        else:
            # Instantiate GUI server and update GUI with port details
            self.gui_service = Service()
            self.gui_service.assign_module(module=self.main_window)
            self.gui_service.assign_logger(logger=self.gui_logger)
            if self.gui_port is None:
                self.gui_server, self.gui_port = create_server(
                    self.gui_service,
                    logger=self.gui_logger,
                    host=socket.gethostbyname(socket.gethostname())
                    )
            else:
                try:
                    self.gui_server = GenericServer(
                        service=self.gui_service,
                        host='localhost',
                        port=self.gui_port
                    )
                except ConnectionRefusedError:
                    self.gui_logger.error(f'Failed to instantiate GUI Server at port {self.gui_port}')
                    raise
            self.gui_server.start()
            self.gui_logger.update_data(data=dict(port=self.gui_port))
            # Update internal attributes and add to list of log clients
            self.client_list[self.GUI_NAME] = QtWidgets.QListWidgetItem(self.GUI_NAME)
            self.port_list[self.GUI_NAME] = [port for port in self.log_server._server.clients][0]
            self.main_window.client_list.addItem(self.client_list[self.GUI_NAME])
            self.client_list[self.GUI_NAME].setToolTip(dict_to_str(self.log_service.client_data[self.GUI_NAME]))
            self.client_data[self.GUI_NAME+module_str] = self.log_service.client_data[self.GUI_NAME]

        self.main_window.gui_label.setText('{} GUI Port: {}'.format(gui_str, self.gui_port))

    def update_terminal(self):
        """ Updates terminal output on GUI """

        to_append = sys.stdout.getvalue()
        self.main_window.terminal.append(to_append)
        self.main_window.buffer_terminal.append(to_append)
        try:
            self.main_window.terminal.moveCursor(QtGui.QTextCursor.End)
        except TypeError:
            pass
        sys.stdout.truncate(0)
        sys.stdout.seek(0)

        # Update buffer terminal
        buffer_str = f'!~{self.update_index}~!{to_append}'
        self.main_window.buffer_terminal.append(buffer_str)
        self.update_index += 1

    def check_disconnection(self):
        """ Checks if a client has disconnected and raises a flag if so"""

        if 'Client disconnected' in sys.stdout.getvalue() or self.disconnection:
            self.disconnection = True

    def disconnect(self):
        """ Handles case where client has disconnected """

        to_del = [client for (client, port) in self.port_list.items() if port not in self.log_server._server.clients]
        for client in to_del:
            print('[INFO] {} disconnected at {}'.format(client, time.strftime("%Y-%m-%d, %H:%M:%S", time.gmtime())))
            self.main_window.client_list.takeItem(self.main_window.client_list.row(self.client_list[client]))
            del self.client_list[client]
            del self.port_list[client]
            del self.log_service.client_data[client]
            del self.client_data[client]
            self.disconnection = False

    def update_connection(self):
        """ Checks if new/updated connections have been made and updates accordingly"""

        # Figure out ports/clients to add
        port_to_add = [port for port in self.log_server._server.clients if port not in self.port_list.values()]
        client_to_add = [client for client in self.log_service.client_data if client not in self.client_list]

        # Add client and update relevant directories + GUI
        if len(client_to_add) > 0:
            client = client_to_add[0]
            self.client_list[client] = QtWidgets.QListWidgetItem(client)
            self.main_window.client_list.addItem(self.client_list[client])
            try:
                self.main_window.client_list.moveCursor(QtGui.QTextCursor.End)
            except TypeError:
                pass
            self.client_list[client].setToolTip(dict_to_str(self.log_service.client_data[client]))
            if len(port_to_add) > 0:
                self.port_list[client] = port_to_add[0]
            self.client_data[client] = self.log_service.client_data[client]

        # Check for updates to client data
        while len(self.log_service.data_updated) > 0:
            self.client_list[self.log_service.data_updated[0]].setToolTip(
                dict_to_str(self.log_service.client_data[self.log_service.data_updated[0]])
            )
            del self.log_service.data_updated[0]

    def start_logger(self):
        """ Starts the log server """

        self.log_service = LogService()
        if self.LOG_PORT is None:
            self.log_server, self.log_port = create_server(
                self.log_service,
                host=socket.gethostbyname(socket.gethostname())
                )
        else:
            try:
                self.log_server = GenericServer(
                    service=self.log_service,
                    host=socket.gethostname(socket.gethostname()),
                    port=self.LOG_PORT
                    )
            except ConnectionRefusedError:
                print(f'Failed to insantiate Log Server at port {self.LOG_PORT}')
                raise
        self.log_server.start()

    def initialize_gui(self):
        """ Initializes basic GUI display """

        # sys.stdout = StringIO()
        # self.app = QtWidgets.QApplication(sys.argv)
        # self.main_window = Window(self.app, gui_template=self.LOGGER_UI)

        ip_str, ip_str_2, log_str = '', '', ''
        if self.proxy:
            self.main_window.setWindowTitle('Launch Control (Proxy)')
            ip_str = 'Master (Local) '
            ip_str_2 = f' ({socket.gethostbyname(socket.gethostname())})'
            log_str = 'Master '
        self.main_window.ip_label.setText(
            f'{ip_str}IP Address: {self.host}'+ip_str_2
            )
        self.main_window.logger_label.setText(f'{log_str} Logger Port: {self.log_port}')

        if self.proxy:
            self.main_window.terminal.setText('Connected to master Log Server. \n')
        self.main_window.terminal.setText('Log messages will be displayed below \n')
        self.main_window.buffer_terminal.document().setMaximumBlockCount(100)

        # Assign widgets for remote access
        self.main_window.assign_container('client_list', 'clients')
        self.main_window.assign_label('buffer_terminal', 'buffer')
        self.main_window.assign_event_button('debug_radio_button', 'debug')

        # Configure list of scripts to run and clicking actions
        self._load_scripts()
        self._configure_clicks()
        self._configure_debug()
        self._configure_debug_combo_select()

        self.main_window.force_update()

    def update_proxy(self):
        """ Updates the proxy with new content using the buffer terminal"""

        # Check clients and update
        self._pull_connections()

        # Get buffer terminal
        buffer_terminal = self.gui_client.get_text('buffer')

        # Parse buffer terminal to get part of the message that is new
        new_msg = buffer_terminal[buffer_terminal.rfind(f'!~{self.update_index+1}~!'):-1]

        # Check if this failed
        if new_msg == '':

            # Check if the buffer is ahead of our last update
            up_str = re.findall(r'!~\d+~!', new_msg)
            if len(up_str) > 0:
                up_in = int(re.findall(r'\d+', up_str[0]))
                if up_in > self.update_index:
                    new_msg = buffer_terminal

        # If we have a new message to add, add it
        if new_msg != '':

            self.main_window.terminal.append(re.sub(r'!~\d+~!', '', new_msg))
            try:
                self.main_window.terminal.moveCursor(QtGui.QTextCursor.End)
            except TypeError:
                pass
            self.update_index = int(re.findall(r'\d+', re.findall(r'!~\d+~!', new_msg)[-1])[0])

    def _configure_clicks(self):
        """ Configures what to do if script is clicked """

        self.main_window.script_list.itemDoubleClicked.connect(self._clicked)

    def _clicked(self):
        """ Launches the script that has been double-clicked

        Opens a new commandline subprocess using subprocess.Popen(bash_cmd) which runs the
        relevant python script, passing all relevant LogClient information via the commandline
        """

        script_to_run = list(self.script_list.keys())[list(self.script_list.values()).index(
            self.main_window.script_list.currentItem()
        )]
        launch_time = time.strftime("%Y-%m-%d, %H:%M:%S", time.gmtime())
        print('Launching {} at {}'.format(script_to_run, launch_time))  # TODO MAKE A LOG STATEMENT

        # Read debug state and set flag value accordingly.

        # Initial configurations: All flags down.
        debug_flag, server_debug_flag, gui_debug_flag = '0', '0', '0'

        # Raise flags if selected in combobox
        if self.debug:
            if self.debug_level == "launcher":
                debug_flag = '1'
            elif self.debug_level == "pylabnet_server":
                server_debug_flag = '1'
            elif self.debug_level == "pylabnet_gui":
                gui_debug_flag = '1'

        # Build the bash command to input all active servers and relevant port numbers to script
        bash_cmd = 'start /min "{}, {}" /wait "{}" "{}" --logip {} --logport {} --numclients {} --debug {} --server_debug {} --gui_debug {}'.format(
            script_to_run,
            launch_time,
            sys.executable,
            os.path.join(os.path.dirname(os.path.realpath(__file__)), script_to_run),
            self.host,
            self.log_port,
            len(self.client_list),
            debug_flag,
            server_debug_flag,
            gui_debug_flag
        )
        client_index = 1
        for client in self.client_list:
            bash_cmd += ' --client{} {} --ip{} {}'.format(
                client_index, remove_spaces(client), client_index, self.client_data[client]['ip']
            )

            # Add port of client's server, if applicable
            if 'port' in self.client_data[client]:
                bash_cmd += ' --port{} {}'.format(client_index, self.client_data[client]['port'])

            # If this client has relevant .ui file, pass this info
            if 'ui' in self.client_data[client]:
                bash_cmd += ' --ui{} {}'.format(client_index, self.client_data[client]['ui'])

            client_index += 1

        # Launch the new process
        subprocess.Popen(bash_cmd, shell=True)

    def _load_scripts(self):
        """ Loads all relevant scripts from current working directory """

        # Get all relevant files
        current_directory = os.path.dirname(os.path.realpath(__file__))
        files = [file for file in os.listdir(current_directory) if (
            os.path.isfile(os.path.join(
                current_directory, file
            )) and '.py' in file and '__init__.py' not in file and 'launch_control.py' not in file and 'launcher.py' not in file
        )]

        for file in files:
            self.script_list[file] = QtWidgets.QListWidgetItem(file.split('.')[0])
            self.main_window.script_list.addItem(self.script_list[file])

    def _copy_master(self):
        """ Updates the GUI to copy the GUI of the master GUI server """

        # Get a dictionary of all client names and tooltip info
        clients = self.gui_client.get_container_info('clients')

        # Update the proxy GUI to reflect the client list of the main GUI
        for client, info in clients.items():
            self.client_list[client] = QtWidgets.QListWidgetItem(client)
            self.main_window.client_list.addItem(self.client_list[client])
            self.client_list[client].setToolTip(info)

            # Add client data
            self.client_data[client] = {}
            if 'ip: ' in info:
                self.client_data[client]['ip'] = info.split('ip: ')[1].split('\n')[0]
            if 'timestamp: ' in info:
                self.client_data[client]['timestamp'] = info.split('timestamp: ')[1].split('\n')[0]
            if 'ui: ' in info:
                self.client_data[client]['ui'] = info.split('ui: ')[1].split('\n')[0]
            if 'port: ' in info:
                self.client_data[client]['port'] = info.split('port: ')[1].split('\n')[0]

    def _pull_connections(self):
        """ Updates the proxy's client list """

        # Get a dictionary of all client names and tooltip info
        clients = self.gui_client.get_container_info('clients')

        # Update the proxy GUI to reflect the client list of the main GUI
        add_clients = list(set(clients.keys()) - set(self.client_list.keys()))
        remove_clients = list(set(self.client_list.keys()) - set(clients.keys()))

        # Add clients
        for client in add_clients:
            self.client_list[client] = QtWidgets.QListWidgetItem(client)
            self.main_window.client_list.addItem(self.client_list[client])
            self.client_list[client].setToolTip(clients[client])

            # Add client data
            self.client_data[client] = {}
            print('Client: '+client)
            if 'ip: ' in clients[client]:
                self.client_data[client]['ip'] = clients[client].split('ip: ')[1].split('\n')[0]
            if 'timestamp: ' in clients[client]:
                self.client_data[client]['timestamp'] = clients[client].split('timestamp: ')[1].split('\n')[0]
            if 'ui: ' in clients[client]:
                self.client_data[client]['ui'] = clients[client].split('ui: ')[1].split('\n')[0]
            if 'port: ' in clients[client]:
                self.client_data[client]['port'] = clients[client].split('port: ')[1].split('\n')[0]

        # Remove clients
        for client in remove_clients:
            self.main_window.client_list.takeItem(self.main_window.client_list.row(self.client_list[client]))
            del self.client_list[client]

    # Defines what to do if debug radio button is clicked.
    def _configure_debug(self):
        self.main_window.debug_radio_button.toggled.connect(self._update_debug_settings)

    # Defines what to do if combobox is changed.
    def _configure_debug_combo_select(self):
        self.main_window.debug_comboBox.currentIndexChanged.connect(self._update_debug_level)

    def _update_debug_settings(self):
        if self.main_window.debug_radio_button.isChecked():
            self.debug = True

            # Enable and show combobox.
            self.main_window.debug_comboBox.setEnabled(True)
            self.main_window.debug_label.setHidden(False)
            self.main_window.debug_comboBox.setHidden(False)

        else:
            self.debug = False
            # Disable and hide combobox.
            self.main_window.debug_comboBox.setEnabled(False)
            self.main_window.debug_label.setHidden(True)
            self.main_window.debug_comboBox.setHidden(True)

        # Update debug level.
        self._update_debug_level(self)

    def _update_debug_level(self, i=0):
        # Set debug level according to combo-box selection.
        # Levels are:
        # pylabnet_server, pylabnet_gui, launcher
        self.debug_level = self.main_window.debug_comboBox.currentText()


def main():
    """ Runs the launch controller """

    log_controller = Controller()
    run(log_controller)

def main_proxy():
    """ Runs the launch controller overriding commandline arguments in proxy mode """

    log_controller = Controller(proxy=True)
    run(log_controller)

def run(log_controller):
    """ Runs the launch controller once a Controller is instantiated"""

    # Instantiate GUI
    if not log_controller.proxy:
        log_controller.start_logger()
    log_controller.initialize_gui()
    log_controller.start_gui_server()

    # Standard operation
    while not log_controller.main_window.stop_button.isChecked():

        # For proxy launch controller, just check main GUI for updates
        if log_controller.proxy:
            log_controller.update_proxy()
        else:

            # Handle external configuration via GUI server
            log_controller.main_window.configure_widgets()
            log_controller.main_window.update_widgets()

            # New terminal input
            if sys.stdout.getvalue() != '':

                # Check for disconnection events
                log_controller.check_disconnection()

                # Handle new connections
                log_controller.update_connection()

                # Update terminal
                log_controller.update_terminal()

            # Handle disconnection events
            if log_controller.disconnection:
                log_controller.disconnect()

        # Update display
        log_controller.main_window.force_update()

    # Exit app (does not close servers)
    sys.exit(log_controller.app.exec_())


if __name__ == '__main__':
    main()



import ctypes

from pylabnet.hardware.interface.wavemeter import WavemeterInterface, WavemeterError
from pylabnet.utils.logging.logger import LogHandler


class Driver(WavemeterInterface):
    """ Hardware class to control High Finesse Wavemeter."""

    def __init__(self, logger=None):
        """ Instantiate wavemeter

        :param logger: instance of LogClient class (optional)
        """

        # Log
        self.log = LogHandler(logger=logger)

        # Load WLM DLL
        try:
            self._wavemeterdll = ctypes.windll.LoadLibrary('wlmData.dll')

        except:
            msg_str = 'High-Finesse WS7 Wavemeter is not properly installed on this computer'
            self.log.error(msg_str)
            raise WavemeterError(msg_str)

        # Set all DLL function parameters and return value types
        self._wavemeterdll.GetWLMVersion.restype = ctypes.c_long
        self._wavemeterdll.GetWLMVersion.argtype = ctypes.c_long

        self._wavemeterdll.GetWLMCount.restype = ctypes.c_long
        self._wavemeterdll.GetWLMCount.argtype = ctypes.c_long

        self._wavemeterdll.GetWavelengthNum.restype = ctypes.c_double
        self._wavemeterdll.GetWavelengthNum.argtypes = [ctypes.c_long, ctypes.c_double]

        self._wavemeterdll.GetFrequencyNum.restype = ctypes.c_double
        self._wavemeterdll.GetFrequencyNum.argtypes = [ctypes.c_long, ctypes.c_double]

        # Check that WLM is running and log details
        self._is_running = self._wavemeterdll.GetWLMCount(0)
        if self._is_running > 0:
            self._wlm_version = self._wavemeterdll.GetWLMVersion(0)
            self.log.info(
                'Connected to High-Finesse Wavemeter WS-{0}'.format(self._wlm_version)
            )

        else:
            msg_str = 'High-Finesse WS7 Wavemeter software not running.\n'
            'Please run the Wavemeter software and try again.'
            self.log.error(msg_str)
            raise WavemeterError(msg_str)

    def get_wavelength(self, channel=1, units='Frequency (THz)'):
        """ Returns the wavelength in specified units for a given channel

        :param channel: Channel number from 1-8
        :param units: "Frequency (THz)" or "Wavelength (nm)". Defaults to frequency.
        """

        if units == 'Wavelength (nm)':
            return self._wavemeterdll.GetWavelengthNum(channel, 0)

        else:
            return self._wavemeterdll.GetFrequencyNum(channel, 0)



import pickle

from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase


class Service(ServiceBase):

    def exposed_set_ao_voltage(self, ao_channel, voltage_pickle):
        voltages = pickle.loads(voltage_pickle)
        return self._module.set_ao_voltage(
            ao_channel=ao_channel,
            voltages=voltages
        )


class Client(ClientBase):

    def set_ao_voltage(self, ao_channel, voltages):
        voltage_pickle = pickle.dumps(voltages)
        return self._service.exposed_set_ao_voltage(
            ao_channel=ao_channel,
            voltage_pickle=voltage_pickle
        )



# -*- coding: utf-8 -*-

"""
This file contains wrappers for Jupyter Notebook Widgets used for controlling the SC20 Shutter
"""

import ipywidgets as widgets


class FilterwheelToogle:
    """ This Class implements a toggle change the filter of a Thorlabs FW102C Filterwheel
    """

    def __init__(self, filterwheel_client):

        # retrieve shutter name
        self.name = filterwheel_client.get_name()

        # retrieve shutter client
        self.filterwheel_client = filterwheel_client

        # retrieve filterwheel dictionary
        self.filters = filterwheel_client.get_filter_dict()

        # retrieve filter names for dropdown selection
        dropdown_options = [(self.filters[i], int(i)) for i in list(self.filters.keys())]

        # Retrieve initial position of filterwheel
        init_pos = filterwheel_client.get_pos()

        # Dropdown Widget
        self.dropdown_raw = widgets.Dropdown(
            options=dropdown_options,
            value=int(init_pos),
        )

        # Add label to toogle
        self.dropdown = widgets.HBox(
            [
                widgets.Label("Filterwheel {name}".format(name=self.name)),
                self.dropdown_raw
            ]
        )

        self.output = widgets.Output()

        # Display Toogle
        display(self.dropdown, self.output)

    def on_value_change(self, change, protect_shutter_client=None):
        with self.output:
            return self.filterwheel_client.change_filter(
                int(change['new']),
                protect_shutter_client
            )



import io
import functools
from contextlib import redirect_stdout


def log_standard_output(func):
    """ Decorator reading standard output produced by function and logging it in LogHandler

    This functions can be used to capture diagnostic print output produced by
    third party APIs and route this to a LogHandler instance. Must be called by
    instance of a hardware class which has a attached self.log LogHandler instance.

    Note that the standard output will only be logged after the function has been called.
    """
    @functools.wraps(func)
    def wrapper(self, *args, **kwargs):

        with io.StringIO() as buf, redirect_stdout(buf):
            # Execute function.
            func_return = func(self, *args, **kwargs)

            # Read standard output from buffer.
            std_output = buf.getvalue()

        # Log every line
        for line in std_output.splitlines():
            self.log.info(line)

        return func_return

    return wrapper




# Historical test module
#
# import rpyc
# import numpy as np
# import pickle
# from pylabnet.utils.logging.logger import LogClient


# class RNG:
#     def __init__(self, mean=0, amp=1, log_host=None, log_port=None, log_tag='RNG', log_level='INFO'):
#         self._mean = mean
#         self._amp = amp

#         self.test_array = np.zeros(1)
#         self.test_array_dict = dict()

#         self.log = LogClient(
#             host=log_host,
#             port=log_port,
#             module_tag=log_tag,
#             level_str=log_level
#         )

#         self.log.info('Started RNG: mean={0}, amp={1}'.format(self._mean, self._amp))

#     def get_params(self):
#         param_dict = dict(mean=self._mean, amp=self._amp)
#         return param_dict

#     def set_params(self, mean=None, amp=None):
#         self.log.info('set_params(mean={}, amp={}) called'.format(mean, amp))

#         if mean is not None:
#             self._mean = mean

#         if amp is not None:
#             self._amp = amp

#         return 0

#     def get_value(self, size=None):
#         values = self._mean + self._amp*(np.random.random(size=size) - 0.5)
#         return values

#     # Timing test
#     def generate_test_array(self, size):
#         self.test_array = np.random.random(size=size)
#         self.log.info(
#             'Successfully generated test_array with size={}'.format(size)
#         )

#     def get_test_array(self):
#         return self.test_array

#     # def get_test_array_pickle(self):
#     #     return pickle.dumps(self.test_array)
#     #
#     # def get_test_array_tostring(self):
#     #     return self.test_array.tostring()

#     # Many active clients test
#     def build_test_array(self, client_number, size=1000):
#         test_ar = client_number + (np.random.random(size=size) - 0.5)
#         self.test_array_dict[str(client_number)] = test_ar

#         self.log.info(
#             'Successfully built test_array. \n'
#             'client_number={}, size={}'.format(client_number, size)
#         )

#         return 0

#     def ret_test_array(self, client_number):
#         return self.test_array_dict[str(client_number)]

#     # Multi-level traceback/stability test
#     def divide(self, a, b):
#         try:
#             c = a / b
#             return c
#         except Exception as exc_obj:
#             self.log.exception('divide(a={}, b={})'.format(a, b))
#             raise exc_obj


# class RNGService(rpyc.Service):

#     _module = None
#     log = None

#     def on_connect(self, conn):
#         # code that runs when a connection is created
#         # (to init the service, if needed)
#         self.log.info('Client connected')

#     def on_disconnect(self, conn):
#         # code that runs after the connection has already closed
#         # (to finalize the service, if needed)
#         self.log.info('Client disconnected')

#     def assign_module(self, module):
#         self._module = module
#         self.log = module.log

#     # ---- Interface --------

#     def exposed_get_params(self):
#         res_dict = self._module.get_params()
#         return pickle.dumps(res_dict)

#     def exposed_set_params(self, mean=None, amp=None):
#         return self._module.set_params(mean=mean, amp=amp)

#     def exposed_get_value(self, size=None):
#         res_ar = self._module.get_value(size=size)
#         return pickle.dumps(res_ar)

#     # Timing test
#     def exposed_generate_test_array(self, size):
#         return self._module.generate_test_array(size=size)

#     def exposed_get_test_array(self):
#         return pickle.dumps(self._module.test_array)

#     # Many active clients test
#     def exposed_build_test_array(self, client_number, size=1000):
#         return self._module.build_test_array(
#             client_number=client_number,
#             size=size
#         )

#     def exposed_ret_test_array(self, client_number):
#         res = self._module.ret_test_array(client_number=client_number)
#         return pickle.dumps(res)

#     # Multi-level traceback/stability test
#     def exposed_divide(self, a, b):
#         return self._module.divide(a=a, b=b)


# class RNGClient:
#     def __init__(self, host, port):

#         # Internal vars to store server info
#         self._host = ''
#         self._port = 0

#         # Internal vars to store refs to server
#         self._connection = None
#         self._service = None

#         # Connect to server
#         self.connect(host=host, port=port)

#     def connect(self, host='place_holder', port=-1):

#         # Update server address if new values are given
#         if host != 'place_holder':
#             self._host = host
#         if port != -1:
#             self._port = port

#         # Clean-up old connection if it exists
#         if self._connection is not None or self._service is not None:
#             try:
#                 self._connection.close()
#             except:
#                 pass

#             self._connection = None
#             self._service = None

#         # Connect to server
#         try:
#             self._connection = rpyc.connect(
#                 host=self._host,
#                 port=self._port,
#                 config={'allow_public_attrs': True}
#             )
#             self._service = self._connection.root

#             return 0

#         except Exception as exc_obj:
#             self._connection = None
#             self._service = None
#             raise exc_obj

#     # Interface methods

#     def get_params(self):
#         res_pickle = self._service.exposed_get_params()
#         return pickle.loads(res_pickle)

#     def set_params(self, mean=None, amp=None):
#         ret_code = self._service.exposed_set_params(
#             mean=mean,
#             amp=amp
#         )
#         return ret_code

#     def get_value(self, size=None):
#         pickled_ret_ar = self._service.exposed_get_value(size=size)
#         return pickle.loads(pickled_ret_ar)

#     # Timing test
#     def generate_test_array(self, size):
#         return self._service.exposed_generate_test_array(
#             size=size
#         )

#     def get_test_array(self):
#         pickled_ret_ar = self._service.exposed_get_test_array()
#         return pickle.loads(pickled_ret_ar)

#     # Many active clients test
#     def build_test_array(self, client_number, size=1000):
#         return self._service.exposed_build_test_array(
#             client_number=client_number,
#             size=size
#         )

#     def ret_test_array(self, client_number):
#         res_pickle = self._service.exposed_ret_test_array(
#             client_number=client_number
#         )
#         return pickle.loads(res_pickle)

#     # Multi-level traceback/stability test
#     def divide(self, a, b):
#         return self._service.exposed_divide(
#             a=a,
#             b=b
#         )



import TimeTagger as TT
import time
import copy
import pickle
from pylabnet.utils.logging.logger import LogHandler
from pylabnet.hardware.interface.gated_ctr import CtrError
from pylabnet.network.core.service_base import ServiceBase
from pylabnet.network.core.client_base import ClientBase


class Wrap:

    def __init__(self, tagger, click_ch, start_ch, logger=None):
        """Instantiate gated counter

        :param tagger: instance of TimeTagger class
        :param click_ch: (int|list of int) clicks on all specified channels
                                    will be summed into one logical channel
        :param start_ch: (int) start trigger channel number
        """

        # Log
        self.log = LogHandler(logger=logger)

        # Reference to tagger
        self._tagger = tagger

        # Log device ID information to demonstrate that connection indeed works
        serial = self._tagger.getSerial()
        model = self._tagger.getModel()
        self.log.info(
            'Got reference to Swabian Instruments TimeTagger device \n'
            'Serial number: {0}, Model: {1}'
            ''.format(serial, model)
        )

        # Gated Counter
        # reference to the TT.CountBetweenMarkers measurement instance
        self._ctr = None
        # number of count bins:
        #   length of returned 1D count array, the size of allocated memory buffer.
        # must be given as argument of init_ctr() call
        self._bin_n = 0

        # Channel assignments
        self._click_ch = 0
        self._start_ch = 0
        # reference to Combiner object
        #   (if _click_ch is a list - then counts on all channels are summed
        #   into virtual channel - self._combiner.getChannel())
        self._combiner = None
        # apply channel assignment
        self.set_ch_assignment(
            click_ch=click_ch,
            start_ch=start_ch
        )

    # ---------------- Interface ---------------------------

    def activate_interface(self):
        return 0

    def init_ctr(self, bin_n, bin_w):

        bin_w = int(bin_w / 1e-12)

        # Close existing counter, if it was initialized before
        self.close_ctr()

        # Instantiate counter measurement
        try:
            self._ctr = TT.TimeDifferences(
                tagger=self._tagger,
                click_channel=self._click_ch,
                start_channel=self._start_ch,
                n_bins=bin_n,
                binwidth=bin_w
            )

            # save bin_number in internal variable
            self._bin_n = bin_n

        # handle NotImplementedError (typical error, produced by TT functions)
        except NotImplementedError:
            # remove reference to the counter measurement
            self._ctr = None

            msg_str = 'init_ctr(): instantiation of TimeDifferences measurement failed'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

        # Prepare counter to be started by start_counting()
        # (TimeDifferences measurement starts running immediately after instantiation,
        # so it is necessary to stop it and erase all counts collected between instantiation and stop() call)
        self._ctr.stop()
        self._ctr.clear()

        return 0

    def close_ctr(self):

        # Try to stop and to clear TT.CountBetweenMarkers measurement instance
        try:
            self._ctr.stop()
            self._ctr.clear()
        except:
            pass

        # Remove reference, set status to "void"
        self._ctr = None

        return 0

    def start_counting(self):

        # Try stopping and restarting counter measurement
        try:
            self._ctr.stop()  # does not fail even if the measurement is not running
            self._ctr.clear()
            self._ctr.start()

            # Wait until the counter is actually ready to count
            time.sleep(0.1)

            return 0

        # handle exception in TT function calls [NotImplementedError]
        except NotImplementedError:
            # Since stop() and clear() methods are very robust,
            # this part is only executed if counter is totally broken.
            # In this case it makes sense to close counter.
            self.close_ctr()

            msg_str = 'start_counting(): call failed. Counter was closed. \n'\
                      'Re-initialize counter by calling init_ctr() again'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

    def stop_counting(self):

        # Try stopping counter measurement
        try:
            # stop counter
            self._ctr.stop()

            return 0

        # handle exception in TT.stop()/TT.clear()
        except NotImplementedError:
            # Since stop() and clear() methods are very robust,
            # this part is only executed if counter is totally broken.
            # In this case it makes sense to close counter.
            self.close_ctr()

            msg_str = 'terminate_counting(): call failed. Counter was closed. \n' \
                      'Re-initialize it by calling init_ctr()'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

    def get_count_trace(self):
        return self._ctr.getData()[0]

    # ------------------------------------------------------

    def get_ch_assignment(self):
        """Returns dictionary containing current channel assignment:
            {
                'click_ch': (int) click_channel_number_including_edge_sign
                'gate_ch': (int) gate_channel_number_including_edge_sign
            }

        :return: dict('click_ch': _, 'gate_ch': _)
        """

        click_ch = copy.deepcopy(self._click_ch)
        start_ch = copy.deepcopy(self._start_ch)

        return dict(click_ch=click_ch, start_ch=start_ch)

    def set_ch_assignment(self, click_ch=None, start_ch=None):
        """Sets click channel and and gate channel.

        This method only changes internal variables
        self._click_ch and self._gate_ch.
        To apply the channel update, call  init_ctr() again.


        :param click_ch: (int|list of int) click channel number
                              positive/negative values - rising/falling edge detection
                              if list is given, clicks on all specified channels
                              will be merged into one logic channel

        :param start_ch: (int) channel number
                             positive/negative - count during high/low gate level

        :return: (dict) actually channel assignment:
                        {
                            'click_channel': (int) click_chnl_num,
                            'gate_channel': (int) gate_chnl_num
                        }
        """

        if click_ch is not None:
            # for convenience bring int type of input to list of int
            if isinstance(click_ch, list):
                click_ch_list = click_ch
            elif isinstance(click_ch, int):
                click_ch_list = [click_ch]
            else:
                # unknown input type
                msg_str = 'set_ch_assignment(click_ch={0}): invalid argument type'\
                          ''.format(click_ch)
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

            # sanity check: all requested channels are available on the device
            all_chs = self.get_all_chs()
            for channel in click_ch_list:
                if channel not in all_chs:
                    msg_str = 'set_ch_assignment(): '\
                              'click_ch={0} - this channel is not available on the device'\
                              ''.format(click_ch)
                    self.log.error(msg_str=msg_str)
                    raise CtrError(msg_str)

            # If several channel numbers were passed, create virtual Combiner channel
            if len(click_ch_list) > 1:
                self._combiner = TT.Combiner(
                    tagger=self._tagger,
                    channels=click_ch_list
                )
                # Obtain int channel number for the virtual channel
                click_ch_list = [self._combiner.getChannel()]

            # Set new value for click channel
            self._click_ch = int(click_ch_list[0])

        if start_ch is not None:

            # sanity check: channel is available on the device
            if start_ch not in self.get_all_chs():
                msg_str = 'set_ch_assignment(): '\
                          'start_ch={0} - this channel is not available on the device'\
                          ''.format(start_ch)
                self.log.error(msg_str=msg_str)
                raise CtrError(msg_str)

            # Set new value for gate channel
            self._start_ch = int(start_ch)

        return self.get_ch_assignment()

    def get_all_chs(self):
        """Returns list of all channels available on the device,
        including edge type sign.

        Positive/negative numbers correspond to detection of rising/falling edges.
        For example:
            1 means 'rising edge on connector 1'
            -1 means 'falling edge on connector 1


        :return: (list of int) list of channel numbers including edge sign.
                Example: [-8, -7, -6, -5, -4, -3, -2, -1, 1, 2, 3, 4, 5, 6, 7, 8]
                Empty list is returned in the case of error.
        """

        # Sanity check: check that connection to the device was established
        if self._tagger is None:
            msg_str = 'get_all_chs(): not connected to the device yet'
            self.log.error(msg_str=msg_str)
            raise CtrError(msg_str)

        channel_list = list(
            self._tagger.getChannelList(TT.TT_CHANNEL_RISING_AND_FALLING_EDGES)
        )
        return channel_list


class Service(ServiceBase):

    def exposed_activate_interface(self):
        return self._module.activate_interface()

    def exposed_init_ctr(self, bin_n, bin_w):
        return self._module.init_ctr(
            bin_n=bin_n,
            bin_w=bin_w
        )

    def exposed_close_ctr(self):
        return self._module.close_ctr()

    def exposed_start_counting(self):
        return self._module.start_counting()

    def exposed_stop_counting(self):
        return self._module.stop_counting()

    def exposed_get_count_trace(self):
        res = self._module.get_count_trace()
        return pickle.dumps(res)


class Client(ClientBase):

    def activate_interface(self):
        return self._service.exposed_activate_interface()

    def init_ctr(self, bin_n, bin_w):
        return self._service.exposed_init_ctr(
            bin_n=bin_n,
            bin_w=bin_w
        )

    def close_ctr(self):
        return self._service.exposed_close_ctr()

    def start_counting(self):
        return self._service.exposed_start_counting()

    def stop_counting(self):
        return self._service.exposed_stop_counting()

    def get_count_trace(self):
        res_pickle = self._service.exposed_get_count_trace()
        return pickle.loads(res_pickle)
